#ifndef _DYNAMICS_H_
#define _DYNAMICS_H_

#include "vec.h"
#include "spring.h"
#include "mesh.h"


class Dynamics
{
public:
	// balls
	std::vector<Vec> positions; 
	std::vector<Vec> velocities; 
	std::vector<float> masses; 
	std::vector<float> invMasses; 
	std::vector<float> radiuses;
	
	// force to draw (used to display collision model but not mechanical model)
	std::vector<bool> forceDraw;
	
	// ground
	Vec groundPosition;
	Vec groundVelocity;
	Vec groundNormal;

	// forces
	std::vector<Spring> springs;
	Vec gravity;
	float viscosity;
	
public:

	// initialize dynamic animation
	Dynamics()
		: groundPosition(0.0f, 0.0f, 0.0f),
		  groundVelocity(0.0f, 0.0f, 0.0f),
		  groundNormal(0.0f, 1.0f, 0.0f),
		  gravity(0.0f, -10.0f, 0.0f), viscosity(1.0f)
		  {}
	
	~Dynamics(){}

	// ACCESSORS
	
	// add a ball to dynamic animation and send back an identifier for this ball
	unsigned int addBall(const Vec& ballPos, const Vec& velocity, float mass, float radius);
	// add a spring between two balls
	unsigned int addSpring(unsigned int ball1, unsigned int ball2, float stiffness, float initLength, float dampingFactor);

	// set gravity vector
	void setGravity(const Vec& grav);
	// force the position of a ball
	void setBallPos(unsigned int ball, const Vec& pos);
	
	// get ball position
	const Vec& getBallPos(unsigned int ball) const {return positions[ball];}
	// get ball radius
	double getBallRadius(unsigned int ball) const {return radiuses[ball];}
	// get number of springs
	unsigned int getNbSprings() const {return (unsigned int) springs.size();}
	// get number of balls
	unsigned int getNbBalls() const {return (unsigned int) positions.size();}
	// get a spring
	const Spring& getSpring (unsigned int spring) const {return springs[spring];}

	// METHODS

	// compute collision between a ball and a plane
	void collisionBallPlane(Vec& x1, Vec& v1, float r1, float invm1,
													Vec& x2, Vec& v2, Vec& normal, float invm2,
													float rebound );

	// compute collision between a ball and a ball
	void collisionBallBall(Vec& x1, Vec& v1, float r1, float invm1,
												 Vec& x2, Vec& v2, float r2, float invm2,
												 float rebound );
	
	// update positions and velocities of dynamic objects after a time step dt
	void animate(float dt);

	// set dynamics from obj file :
	void setFromMesh(const char* file, Mesh *mesh);

	// add a spherical collision model that will be displayed no matter what :
	void addSphereModel(const char* file);

};

#endif // _DYNAMICS_H_
