#ifdef WIN32
#include <windows.h>
#endif

//GLEW for OpenGL extentions
#include <GL/glew.h>

#include <GL/glut.h>
#include "glm/glm.hpp"
#include "glm/gtc/type_ptr.hpp"
#include "glm/gtc/matrix_transform.hpp"

//Standard suff
#include <cstdio>
#include <iostream>
#include <string>
#include <assert.h>
#include <cmath>
#ifndef M_PI
	#define M_PI 3.14159265358979323846f
#endif
#include <vector>

#include "animation.h"

using namespace std;

bool draw_axes = false;

Animation anim;

const char *windowTitle = "Character animation";

/* Events (mouse + modifiers) */
#define MOUSE_SHIFT  0x0001
#define MOUSE_CTRL   0x0002
#define MOUSE_ALT    0x0004
#define MOUSE_LEFT   0x0010
#define MOUSE_MIDDLE 0x0020
#define MOUSE_RIGHT  0x0040

struct CMouse
{
    double	X ,Y,             /* Normalized mouse position     */
    X0,Y0,			 /* Previous position            */
    Dx,Dy;			 /* Displacement                      */
    int	Left, Middle, Right;  /* Mouse buttions state  */
    int	Shift, Ctrl, Alt;     /* Modifiers state */
    int	State;              /* Global state (see following flags) */
};
// Flags for mouse state
#define SHIFT  0x0001
#define CTRL   0x0002
#define ALT    0x0004
#define LEFT   0x0010
#define MIDDLE 0x0020
#define RIGHT  0x0040


struct CView
{
    double	Distance;
    double	AngleY;
    double	AngleX;

	double PosX;
	double PosY;

    bool	Wireframe;
};


CView		MyView;
CMouse		MyMouse;
int WindowW = 512, WindowH = 512;

void Init();
void ReloadShader();
#define BUFFER_OFFSET(a) ((char*)NULL + (a))
char* ProgramName;


///////////////////////////////////////////////////////////////
//Shader
GLuint MyProgram;                  // Shader handle
char* ShaderNoLighting = (char*) "skinningShader"; // Shader file names
char* Shader=(char*) "";

typedef glm::vec4 Point;
typedef glm::vec3 Normal;
typedef glm::mat4 Mat4;

//////////////////////////////////////////////////////////////
// Points
unsigned int num_points;			// number of vertices in skinning
Point *initial_points;				// initial positions of vertices (ones being fed to shader)
GLuint initial_points_vboid;		// buffer object
//////////////////////////////////////////////////////////////
// Normals
Normal *initial_normals;			// initial normals of vertices (ones being fed to shader)
GLuint initial_normals_vboid;		// buffer object
//////////////////////////////////////////////////////////////
// Triangles
unsigned num_triangle_indices;		// number of triangle indices = 3* number of triangles
GLuint triangles_ibo;				// buffer object
//////////////////////////////////////////////////////////////
// Bones
unsigned int num_bones;				// number of bones for skinning
const unsigned int max_bones = 250;	// Must be the same as in the shader
const unsigned int m_origin_size = max_bones * sizeof(Mat4);
Mat4 matrices[max_bones];			// current transformation of joints
Mat4 inv_matrices[max_bones];		// inverse of initial transformation of joints (at rest pose)
Mat4 m_origin[max_bones];			// transformation of bones from time 0 to time t
//////////////////////////////////////////////////////////////
// bone_weight
const unsigned num_frames_per_point = 4;	// number of bones influencing each vertex (same for all vertices)
float *bone_weight;							// skinning weights (see skinning.h)
float *bone_index;							// bone indices (see skinning.h)
GLuint bone_weight_vboid;					// buffer object
GLint bone_weight_attrib_location;			// weight id in shader
GLuint bone_index_vboid;					// buffer object
GLint bone_index_attrib_location;			// id in shader


template<typename T>
inline void create_buffer_object( GLuint* vbo, unsigned long BUFFER_TYPE, unsigned long ACCESS_TYPE, T* data, unsigned num_variables )
{
    glGenBuffers(1, vbo);
    glBindBuffer(BUFFER_TYPE, *vbo);
    unsigned int vbo_size = num_variables * sizeof(T);
    glBufferData(BUFFER_TYPE, vbo_size, 0, ACCESS_TYPE);
    glBufferSubData(BUFFER_TYPE, 0, vbo_size, data);
    glBindBuffer(BUFFER_TYPE, 0);
}


void drawAxes()
{
    glBegin(GL_LINES);
    glColor3f(1,0,0);
    glVertex3f(0,0,0);
    glVertex3f(1,0,0);
    glColor3f(0,1,0);
    glVertex3f(0,0,0);
    glVertex3f(0,1,0);
    glColor3f(0,0,1);
    glVertex3f(0,0,0);
    glVertex3f(0,0,1);
    glEnd();
}

void init_matrices(vector<glm::vec3> _posVec, vector<glm::vec3> _RotAxis, vector<float> _AngleVec) 
{
	// Compute inv_matrices
	for( unsigned int i=0; i<num_bones; i++ ) {
        matrices[i]= glm::translate(glm::mat4(1.0f), _posVec[i]);
        matrices[i]= glm::rotate( matrices[i], _AngleVec[i], _RotAxis[i]);    
		inv_matrices[i] = glm::inverse(matrices[i]);
    }
}
void update_matrices(vector<glm::vec3> _posVec, vector<glm::vec3> _RotAxis, vector<float> _AngleVec)
{
	// Update matrices with (_posVec, _RotAxis, _AngleVec) and compute m_origin
	for( unsigned int i=0; i<num_bones; i++ ) {
      	matrices[i]= glm::translate(glm::mat4(1.0f), _posVec[i]);
        matrices[i]= glm::rotate( matrices[i], _AngleVec[i], _RotAxis[i]);    
		if (anim.skinning.parent[i]>=0) matrices[i] = m_origin[anim.skinning.parent[i]]*matrices[i];
		m_origin[i] = matrices[i]*inv_matrices[i];
    }
}

void animate()
{
	// Set time step :
	anim.setDt();

	// Animate with skinning :
	//anim.animateSkinning();
	// Update the bone information passed to vertex buffer object :
	update_matrices(anim.skinning.jointsPos, anim.skinning.jointsAxis, anim.skinning.jointsAngle);
	// Update the collision model with transformation of bones :
	anim.updateCollisionModelFromSkinning(m_origin);
	// Compute the dynamics :
	anim.animateDynamics();
}

void Display()
{

	//Standard stuff
	glClearColor(0., 0., 0., 1);
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0, double( WindowW )/double( WindowH ), .1, 10000 );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glTranslatef( MyView.PosX, MyView.PosY, -MyView.Distance );
	glRotatef( MyView.AngleX, 1, 0, 0 );
	glRotatef( MyView.AngleY, 0, 1, 0 );

	glDisable( GL_LIGHTING );
	if( draw_axes ) {
		for(unsigned int i=0; i<num_bones; i++ ) {
			glPushMatrix( );
			glMultMatrixf( glm::value_ptr(matrices[i]) );
			drawAxes();
			glPopMatrix();
		}
	}

	glColor3f( 1,1,1 );


	//////////////////////////////////
	//////////// Animation ///////////
	//////////////////////////////////
	animate();
	
	// Display dynamics and cloth :
	anim.renderScene();

	// Display character :
	// Activate the shader
	glUseProgram( MyProgram );
	GLint shader_matrices = glGetUniformLocation( MyProgram, "matrices" );
	assert( shader_matrices != -1 );
	glUniformMatrix4fv( shader_matrices, num_bones, false, glm::value_ptr(m_origin[0]) );

	// activate weight data
	glBindBuffer(GL_ARRAY_BUFFER, bone_weight_vboid);
	glVertexAttribPointer(bone_weight_attrib_location, num_frames_per_point, GL_FLOAT, GL_FALSE, 0, 0 );
	glEnableVertexAttribArray(bone_weight_attrib_location);

	// activate weight index data
	glBindBuffer(GL_ARRAY_BUFFER, bone_index_vboid);
	glVertexAttribPointer(bone_index_attrib_location, num_frames_per_point, GL_FLOAT, GL_FALSE, 0, 0 );
	glEnableVertexAttribArray(bone_index_attrib_location);

	// activate vertex data
	glBindBuffer(GL_ARRAY_BUFFER, initial_points_vboid);
	glVertexPointer(4, GL_FLOAT, 0, BUFFER_OFFSET(0));
	glEnableClientState(GL_VERTEX_ARRAY);

	glBindBuffer(GL_ARRAY_BUFFER, initial_normals_vboid);
	glNormalPointer(GL_FLOAT, 0, 0);
	glEnableClientState(GL_NORMAL_ARRAY);

	// draw
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, triangles_ibo);
	glColor4f(0.75,0.25,0,1);
	glDrawElements(GL_TRIANGLES, num_triangle_indices, GL_UNSIGNED_INT, 0 );
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	

	// deactivate data
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableVertexAttribArray(bone_weight_attrib_location);
	glDisableVertexAttribArray(bone_index_attrib_location);
	glUseProgram( 0 );

	glutSwapBuffers();
}


std::ostream& operator << ( std::ostream& out, const glm::vec3& v ){
    out << v[0] <<" "<< v[1] <<" "<<v[2];
    return out;
}

std::ostream& operator << ( std::ostream& out, const glm::vec4& v ){
    out << v[0] <<" "<< v[1] <<" "<<v[2]<<" "<<v[3];
    return out;
}

float operator * ( const glm::vec3& u, const glm::vec3& v){
    return u[0]*v[0] + u[1]*v[1] + u[2]*v[2];
}

void Init()
{
    glewInit();

    // Load shaders
    ReloadShader();

	anim.initScene(num_frames_per_point);

	// Get mesh info to shader :
	num_points = anim.characterMesh->points.size();
	initial_points = (Point*)calloc(sizeof(Point), num_points);
	initial_normals = (Normal*)calloc(sizeof(Normal), num_points);
	for (unsigned int i = 0 ; i < anim.characterMesh->points.size() ; i++) {
		initial_points[i] = anim.characterMesh->points[i];
		initial_normals[i] = anim.characterMesh->normals[i];
	}
	vector<unsigned int> triangles;
	for (unsigned int i = 0 ; i < anim.characterMesh->triangles.size() ; i++) {
            triangles.push_back(anim.characterMesh->triangles[i]);
    }

	// Get bone info to shader :
	num_bones = anim.skinning.num_bones;
	init_matrices(anim.skinning.jointsInitPos, anim.skinning.jointsInitAxis, anim.skinning.jointsInitAngle);
	assert( num_bones <= max_bones );

	// Get skinning info to shader :
	bone_weight = (float*)calloc(sizeof(float),num_points * num_frames_per_point);
	bone_index = (float*)calloc(sizeof(float),num_points * num_frames_per_point);
	for (unsigned int i = 0 ; i < num_points * num_frames_per_point ; i++) {
            bone_weight[i] = anim.skinning.bone_weight[i];
			bone_index[i] = anim.skinning.bone_index[i];
    }

    glDisable(GL_CULL_FACE);
    glEnable( GL_DEPTH_TEST );

    // create buffer objects
    create_buffer_object( &initial_points_vboid , GL_ARRAY_BUFFER, GL_STATIC_DRAW, initial_points , num_points );
    create_buffer_object( &initial_normals_vboid, GL_ARRAY_BUFFER, GL_STATIC_DRAW, initial_normals, num_points );
    create_buffer_object( &bone_weight_vboid, GL_ARRAY_BUFFER, GL_STATIC_DRAW, bone_weight, num_points * num_frames_per_point );
    create_buffer_object( &bone_index_vboid, GL_ARRAY_BUFFER, GL_STATIC_DRAW, bone_index, num_points * num_frames_per_point );

    num_triangle_indices = triangles.size();
    create_buffer_object( &triangles_ibo, GL_ELEMENT_ARRAY_BUFFER, GL_STATIC_DRAW, &triangles[0], triangles.size()  );
    
    /** Get access to shader data */
    glUseProgram( MyProgram );
    bone_weight_attrib_location = glGetAttribLocation(MyProgram, "weight");
    //assert (bone_weight_attrib_location!=-1 );
    bone_index_attrib_location = glGetAttribLocation(MyProgram, "weightIndex");
    //assert (bone_index_attrib_location!=-1 );
    glUseProgram( 0 );
}


void ReloadShader()
{
    cerr << "-Compiling-------------------------------------" << endl;

    char FragmentShaderFile[1000];
    char VertexShaderFile[1000];

    Shader = ShaderNoLighting ;

    sprintf( FragmentShaderFile, "%s.fs", Shader );
    sprintf( VertexShaderFile, "%s.vs", Shader );

    GLchar* SourceCode = new GLchar[10000];
    FILE* SourceFile;
    GLint SourceCodeLen;
    GLuint MyFragmentShader;
    GLuint MyVertexShader;
    GLchar Log[10000];
    int L;

    //Read fragment shader source
    SourceFile = fopen( FragmentShaderFile, "r" );
    if( !SourceFile )
    {
        cerr << "Could not open " << FragmentShaderFile << " for reading" << endl;
        return;
    }
    SourceCodeLen = (GLint)fread( SourceCode, 1, 10000, SourceFile );
    fclose( SourceFile );

    MyFragmentShader = glCreateShader( GL_FRAGMENT_SHADER );
    const char *FragmentLines = SourceCode;
    glShaderSource( MyFragmentShader, 1, &FragmentLines, &SourceCodeLen );
    //Compile fragment shader
    glCompileShader( MyFragmentShader );
    glGetShaderInfoLog( MyFragmentShader, 10000, &L, Log );
    if( L ) cerr << "Compile log for fragment shader : " << endl << Log << endl;
    else cerr << "Fragment shader Ok" << endl;


    //Read vertex shader source
    SourceFile = fopen( VertexShaderFile, "r" );
    if( !SourceFile )
    {
        cerr << "Could not open " << VertexShaderFile << " for reading" << endl;
        return;
    }
    SourceCodeLen = (GLint)fread( SourceCode, 1, 10000, SourceFile );
    fclose( SourceFile );

    MyVertexShader = glCreateShader( GL_VERTEX_SHADER );
    const char *VertexLines = SourceCode;
    glShaderSource( MyVertexShader, 1, &VertexLines, &SourceCodeLen );
    //Compile vertex shader
    glCompileShader( MyVertexShader );
    glGetShaderInfoLog( MyVertexShader, 10000, &L, Log );
    if( L ) cerr << "Compile log for vertex shader : " << endl << Log << endl;
    else cerr << "Vertex shader Ok" << endl;


    //Link fragment and vertex shaders
    MyProgram = glCreateProgram();
    glAttachShader( MyProgram, MyFragmentShader );
    glAttachShader( MyProgram, MyVertexShader );
    glLinkProgram( MyProgram );
    glGetProgramInfoLog( MyProgram, 10000, &L, Log );
    if( L ) cerr << "Compile log for link : " << endl << Log << endl;
    else cerr << "Link Ok" << endl;
    cerr << flush;
}

////////////////////////////////////////////////////////////////////
//Basic GLUT stuff
////////////////////////////////////////////////////////////////////

void Usage()
{
    cerr << "Usage :" << endl;
    cerr << "\t" << ProgramName << endl;
    cerr << "Current values:" << endl;
    cerr << "\tshader = " << Shader << endl;
    cerr << endl;
    cerr << "Controls:" << endl;
    cerr << "\tLeftMouseDrag = rotate camera" << endl;
    cerr << "\tMiddleMouseDrag = zoom camera" << endl;
    cerr << "\tRightMouseDrag = translate camera" << endl;
    cerr << "\tF5 = reload shaders" << endl;
    cerr << "\tH = this help" << endl;
    cerr << "\tQ = quit" << endl;
}

void Keypressed( unsigned char Key,int /*x*/, int /*y*/ )
{
    switch( Key )
    {
    case 27:
    case 'q':
    case 'Q':
        exit(0);
        break;

    case 'b':
    case 'B':
        draw_axes = !draw_axes ;
        break;

    case 'h':
    case 'H':
        Usage();
        break;


    case 'w':
    case 'W':
        MyView.Wireframe = !MyView.Wireframe;
        break;
    }
}

void SpecialKeypressed( int Key,int /*x*/,int /*y*/ )
{
    switch( Key )
    {
    case GLUT_KEY_F1:
        Usage();
        break;

    case GLUT_KEY_F5:
        ReloadShader();
        break;

    case GLUT_KEY_UP:
        break;

    case GLUT_KEY_DOWN:
        break;

    }
}

void MouseMoved(int x,int y) 
{
    MyMouse.X0 = MyMouse.X;
    MyMouse.Y0 = MyMouse.Y;
    MyMouse.X =          x /(double)WindowW;
    MyMouse.Y = (WindowH-y)/(double)WindowH;
    MyMouse.Dx = MyMouse.X-MyMouse.X0;
    MyMouse.Dy = MyMouse.Y-MyMouse.Y0;
}

void MouseDragged( int x,int y)  
{
    MyMouse.X0 = MyMouse.X;
    MyMouse.Y0 = MyMouse.Y;
    MyMouse.X =          x /(double)WindowW;
    MyMouse.Y = (WindowH-y)/(double)WindowH;
    MyMouse.Dx = MyMouse.X-MyMouse.X0;
    MyMouse.Dy = MyMouse.Y-MyMouse.Y0;

    switch( MyMouse.State )
    {
    case 0:
        break;
    case MOUSE_LEFT:
        MyView.AngleY += MyMouse.Dx*50;
        MyView.AngleX -= MyMouse.Dy*50;
        break;
    case MOUSE_MIDDLE:
        MyView.Distance -= MyMouse.Dy * MyView.Distance;
        break;
    case MOUSE_RIGHT:
		MyView.PosX += MyMouse.Dx*25;
        MyView.PosY += MyMouse.Dy*25;
        break;
    case MOUSE_SHIFT|MOUSE_LEFT:
        break;
    case MOUSE_SHIFT|MOUSE_MIDDLE:
        break;
    case MOUSE_SHIFT|MOUSE_RIGHT:
        break;
    case MOUSE_CTRL|MOUSE_LEFT:
        break;
    case MOUSE_CTRL|MOUSE_MIDDLE:
        break;
    case MOUSE_CTRL|MOUSE_RIGHT:
        break;
    }

    glutPostRedisplay();
}


void MouseClicked( int Button, int State, int x, int y )
{
    int Modifiers;
    switch( Button )
    {
    case GLUT_LEFT_BUTTON:
        MyMouse.Left = ( State == GLUT_DOWN ); break;
    case GLUT_MIDDLE_BUTTON:
        MyMouse.Middle = ( State == GLUT_DOWN ); break;
    case GLUT_RIGHT_BUTTON:
        MyMouse.Right = ( State == GLUT_DOWN );
        break;
    }

    Modifiers = glutGetModifiers();
    MyMouse.Shift = ( Modifiers & GLUT_ACTIVE_SHIFT ) ?1:0;
    MyMouse.Ctrl  = ( Modifiers & GLUT_ACTIVE_CTRL )  ?1:0;
    MyMouse.Alt   = ( Modifiers & GLUT_ACTIVE_ALT )   ?1:0;

    MyMouse.State = MOUSE_SHIFT*MyMouse.Shift | MOUSE_CTRL*MyMouse.Ctrl | MOUSE_ALT*MyMouse.Alt |
                    MOUSE_LEFT*MyMouse.Left | MOUSE_MIDDLE*MyMouse.Middle | MOUSE_RIGHT*MyMouse.Right;


    MyMouse.X0 = MyMouse.X; MyMouse.X = x /(double)WindowW; 		MyMouse.Dx = MyMouse.X0 - MyMouse.X;
    MyMouse.Y0 = MyMouse.Y; MyMouse.Y = (WindowH-y)/(double)WindowH; MyMouse.Dy = MyMouse.Y0 - MyMouse.Y;

    switch( MyMouse.State )
    {
    case 0:
        break;
    case MOUSE_LEFT:
        break;
    case MOUSE_MIDDLE:
        break;
    case MOUSE_RIGHT:
        break;
    case MOUSE_SHIFT|MOUSE_LEFT:
        break;
    case MOUSE_SHIFT|MOUSE_MIDDLE:
        break;
    case MOUSE_SHIFT|MOUSE_RIGHT:
        break;
    case MOUSE_CTRL|MOUSE_LEFT:
        break;
    case MOUSE_CTRL|MOUSE_MIDDLE:
        break;
    case MOUSE_CTRL|MOUSE_RIGHT:
        break;
    }
    glutPostRedisplay();
}


void ReshapeWindow( int Width,int Height ) 
{
    WindowW = Width;
    WindowH = Height;
    glViewport(0,0,WindowW,WindowH);
    glutPostRedisplay();
}


int main(int argc,char **argv)
{
    glutInit(&argc, argv);
    glutInitWindowSize(640, 480);
	glutInitWindowPosition(300,100);
    glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
    glutCreateWindow(windowTitle);
    glutDisplayFunc( Display );
    glutReshapeFunc( ReshapeWindow );
    glutMouseFunc( MouseClicked );
    glutMotionFunc( MouseDragged );
    glutPassiveMotionFunc( MouseMoved );
    glutKeyboardFunc( Keypressed );
    glutSpecialFunc( SpecialKeypressed );
    glutIdleFunc( Display );


    ProgramName = argv[0];


    MyView.Distance = 60.0;
    MyView.AngleY = 0;
    MyView.AngleX = 20;
	MyView.PosX = -0.7;
	MyView.PosY = 0.0;


    Init();
    Usage();
    glutMainLoop();

    return 0;
}

